#ifndef DIFFIE_HELMAN_H_INCLUDED
#define DIFFIE_HELMAN_H_INCLUDED

#include "signature.h"
#include "encrypt_decrypt_packet.h"
#include "montgomery.h"
#include <stdio.h>

static const u32 dh_length_x32 = 128; // 4096 bit, do not change


//--------------------------------
// No dynamic allocations in this class and nested classes
//
class INITIALIZED_SESSION
{
int initialized;
u32 id[fingerprint_length_x32]; // 256 bit ID of remote side
ENCRYPT_DECRYPT_PACKET encryptor;
ENCRYPT_DECRYPT_PACKET decryptor;

public:
     INITIALIZED_SESSION();   
    ~INITIALIZED_SESSION();
     void reset();
     u32 *session_id(); // Returns NULL if not initialized    
     void init(u32 const *session_id, u32 const *encrypt_key, u32 const* decrypt_key);
     u32 encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes);  // Packet encryption, returns length of encrypted block (= src_length + 16)
     u32 decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes); // length           // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)
                   
   
};

// Client Side

class TX_SECURE_SESSION
{
u32 len;    
u32 *x; // secret
u32 *p;
u32 *g;
u32 *gxp; // g^x mod p: open data
u32 *signature;
u32 *far_signature_decoded; 
SIGNATURE *ds;
TEST_GENERATOR_MODULO_P *test_generator_modulo_p;
MONTGOMERY_POW_MOD *mpm;
DIGITAL_SIGNATURE *digital_signature;
SECRET_DATA *dummy_secret_data;

//---- Fixed size ----
int session_is_initialized;
u32 session_id[fingerprint_length_x32];
u32 encryption_key[8]; // 256 bit key
u32 decryption_key[8]; // 256 bit key
//--------------------

public:
    TX_SECURE_SESSION(u32 length_x32 = signature_length_x32);
    ~TX_SECURE_SESSION();

    TX_SECURE_SESSION & operator = (const TX_SECURE_SESSION &);

    void reset();
    int session_is_active(u32 *session_id = NULL);
    
    u32 init_session(u8 *open_data, SECRET_DATA const* secret_data, u8 const *password, const u32 password_length_bytes); // Initiate session; returns number of bytes
    int process_response(u8 const *response_data,  int (*verify_certificate)(SERTIFICATE const*) = NULL); // Returns 1 if successful initialization
    int start_initialized_session(INITIALIZED_SESSION *initialized_session); // If DH exchange was successful, we can start initialized session from it 

    u32 get_user_info(u8* user_info);
};


// Server side
class RX_SECURE_SESSION
{
u32 len;    
u32 *y; // secret Y
u32 *p;
u32 *g;
u32 *gxp; // g^x mod p: open data
u32 *gyp; // g^x mod p: open data
u32 *signature;
u32 *far_signature_decoded;
u8 *dummy_data;

DIGITAL_SIGNATURE *digital_signature;
SIGNATURE *ds;
SERTIFICATE *far_certificate;
MONTGOMERY_POW_MOD *mpm;

//------- fixed size --------
int session_is_initialized;
u32 session_id[fingerprint_length_x32];
u32 encryption_key[8]; // 256 bit key
u32 decryption_key[8]; // 256 bit key

public:
    RX_SECURE_SESSION(u32 length_x32 = signature_length_x32);
    ~RX_SECURE_SESSION();
    void reset();
    int session_is_active(u32 *session_id = NULL);   
    u32 response_to_init_session(u8 *response_data, u8 const *open_data, SECRET_DATA const* secret_data,
                                 u8 const *password, const u32 password_length_bytes,
                                 int (*verify_certificate)(SERTIFICATE const*) = NULL);

    int start_initialized_session(INITIALIZED_SESSION *initialized_session); // If DH exchange was successful, we can start initialized session from it 

	u32 get_user_info(u8* user_info);
};
//------------------------------------------
// EL GAMAL using certificate as open key
//
class EL_GAMAL_MODE
{
u32 *common_secret;
u32 *open_key;    
u32 *y; // secret key
MONTGOMERY_POW_MOD *mpm;
DIGITAL_SIGNATURE *ds;

u32 encryption_key[8]; // 256 bit key    
u32 data_hash[8]; // 256 bit hash
SIGNATURE signature;

KEYSTREAM_GENERATOR keystream;

void prepare_encryption_data(SERTIFICATE const *sert);
void prepare_decryption_data(SECRET_DATA const* secret_data, u8 const * password, const u32 password_length_bytes);

public:
    EL_GAMAL_MODE();
    ~EL_GAMAL_MODE();

    // Returns number of bytes (0 = error)
    // With optional signing of data

    u64 encrypt(FILE *outfile, FILE *infile, 
                SERTIFICATE const *sert,  // destination  
                SECRET_DATA const *secret_data = NULL,  // signature 
                u8 const *password = NULL, const u32 password_length_bytes = 0);

     // Returns number of bytes (0 = error)
     // With optional signing of data

    u32 encrypt(u8 *output_data, void const * input_data, const u32 data_length_bytes, 
                                SERTIFICATE const *sert, SECRET_DATA const *secret_data = NULL, 
                                                                       u8 const *password = NULL, const u32 password_length_bytes = 0);
    
    u64 decrypt(FILE *outfile, FILE *infile, 
                SECRET_DATA const *secret_data, // Recipient key for decryption 
                u8 const *password, const u32 password_length_bytes, // Recipient password
                int (*verify_certificate)(SERTIFICATE const*) = NULL); // Callback for sender verification 

    u32 decrypt(u8 *output_data, void const * input_data, const u32 input_data_length_bytes,
                                              SECRET_DATA const *secret_data, // Recipient key for decryption 
                                              u8 const *password, const u32 password_length_bytes, // Recipient password
                                               int (*verify_certificate)(SERTIFICATE const*) = NULL); // Callback for sender verification 

    void reset();

};

#endif

