// Encryption/Decryption packets

#include "serialize.h"
#include "hash_func.h"
#include "random.h"
#include "encrypt_decrypt_packet.h"
//#include <stdio.h>
#include <string.h>
#include "large_math.h"



ENCRYPT_DECRYPT_PACKET::ENCRYPT_DECRYPT_PACKET(u32 len_bytes)
{
max_data_len = len_bytes;    
ENCRYPT_DECRYPT_PACKET::reset();
}

ENCRYPT_DECRYPT_PACKET::~ENCRYPT_DECRYPT_PACKET()
{
ENCRYPT_DECRYPT_PACKET::reset();
}
//----------------------------------------
// Initialize 128 bit MIC key u32[4] from 256 bit key u32[8]
//
void ENCRYPT_DECRYPT_PACKET::init(u32 const *key)
{
keystream.init_key(key); // Initialize PRNG key

// Initialize MIC key (this is different procedure from initialization of PRNG key)

mic_key[0] = key[3];
mic_key[1] = key[2];
mic_key[2] = key[1];
mic_key[3] = key[0];

for(u32 ci = 0; ci < 8; ci += 2) 
 {
 X_64(mic_key, key + ci);  // 64 bit data; 64 bit hash

 mic_key[2] += mic_key[0];
 mic_key[3] += mic_key[1];
 
 I_64(mic_key + 2, key + ci);  // 64 bit data; 64 bit hash
 
 mic_key[0] += mic_key[2];
 mic_key[1] += mic_key[3];

 I_64(mic_key, key + ci);  // 64 bit data; 64 bit hash
 
 mic_key[2] += mic_key[0];
 mic_key[3] += mic_key[1];
  
 X_64(mic_key + 2, key + ci);  // 64 bit data; 64 bit hash
 }

reset_stack();
}
//=============================================================
// Compute Message Identification Code (to append or to verify)
//
u32 ENCRYPT_DECRYPT_PACKET::compute_mic(u8 const *packet)
{
// Calculate hash over encoded data
// Initialize with mic_key[0,1]

u32 len_bytes = bytes_to_u32(packet + packet_length_offset);
u64 cntr = bytes_to_u64(packet + packet_counter_offset);

u32 tmp_hash[2];

tmp_hash[0] = mic_key[0];
tmp_hash[1] = mic_key[1];

Hash_Block_X_64(tmp_hash, packet, len_bytes + 4 + 8); // include counter + len + data

// Encrypt with mic_key[2,3]

X_64(tmp_hash, mic_key + 2);

// Return 32 bit hash

return (tmp_hash[0] + tmp_hash[1]);
}
// Destroy all secret data

void ENCRYPT_DECRYPT_PACKET::reset()
{
cheap_random_number(mic_key, 4);
keystream.reset();
reset_stack();
}

u32 ENCRYPT_DECRYPT_PACKET::encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes)
{
if(!src_len_bytes) return 0;
    
if(src_len_bytes > max_data_len) return 0;


// Put data length into packet 

u32_to_bytes(dest, src_len_bytes);    

// Put counter into packet    

u64 keystream_cntr = keystream.counter();

u64_to_bytes(dest + packet_counter_offset, keystream_cntr);


// Collect entropy

u32 entropy[2];
zero(entropy, 2);
Hash_Block_X_64(entropy, src, src_len_bytes);
collect_entropy(entropy);
zero(entropy, 2);


// Encode data by XORing with keystream

for(u32 ci = 0; ci < src_len_bytes; ci++)
 {
 dest[ci + packet_data_offset] = src[ci] ^ keystream.get_keystream_byte();
 }

// Calculate hash over encoded data

u32 mic = compute_mic(dest);

// Append hash

u32_to_bytes(dest + packet_data_offset + src_len_bytes, mic);

return (src_len_bytes + packet_data_offset + 4); // added MIC
}

//======================================================================
// Return 0 = invalid packet (out of sequence, corrupt, or invalid MIC)
//
u32 ENCRYPT_DECRYPT_PACKET::decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes)
{

if(src_len_bytes <= packet_overhead_bytes) return 0;

if(src_len_bytes > (packet_overhead_bytes + max_data_len)) return 0;
    
// Get data length 

u32 len = bytes_to_u32(src + packet_length_offset);

if(!len) return 0; // invalid length

if(len > max_data_len) return 0;

if((len + packet_overhead_bytes) > src_len_bytes) return 0;

// Check MIC

u32 mic = compute_mic(src);

u32 mic_ = bytes_to_u32(src + packet_data_offset + len);

if(mic != mic_) return 0; // Invalid MIC


// Get counter    

u64 cntr = bytes_to_u64(src + packet_counter_offset);


u64 keystream_cntr = keystream.counter();

if(cntr < keystream_cntr) return 0; // Invalid packet

if((cntr - keystream_cntr) > max_skip_counter) return 0; // Too much forward
 
if(!keystream.counter(cntr)) return 0; // Invalid cntr

// Decode data by XORing with keystream

for(u32 ci = 0; ci < len; ci++)
 {
 dest[ci] = src[ci + packet_data_offset] ^ keystream.get_keystream_byte();
 }

// Collect entropy

u32 entropy[2];
zero(entropy, 2);
Hash_Block_X_64(entropy, dest, len);
collect_entropy(entropy);
zero(entropy, 2);

return len;
}
    

//======================== <END> ===================
