// Hash function 

#include "types.h"
#include "xtea.h"
#include "idea.h"
#include "hash_func.h"

//----------------------------
// Nonlinear 128-bit mixing
// Minimum 3 rounds over data to achieve good diffusion
// This function mixes key and data in hash compression functions
// to fortify against weak keys and related key attacks
//


void mix_data(u32 *x, const u32 len, const u32 rounds)
{
for(u32 i = 0; i < rounds; i++)
 {
 for(u32 j = 1; j < len; j++) 
  {
  x[j] += k_mix(x[j-1]);
  }
 x[0] += k_mix(x[len-1]); 
 } 
}


void unmix_data(u32 *x, const u32 len, const u32 rounds)
{
for(u32 i = 0; i < rounds; i++)
 {
 x[0] -= k_mix(x[len-1]);
   
 for(u32 j = len-1; j > 0; j--)
  {
  x[j] -= k_mix(x[j-1]);
  }
 }
}






static const u32 mix_128_rounds = 4;
//
static void mix_128(u32 *x)
{

for(u32 i = 0; i < mix_128_rounds; i++)
 {
 x[1] += k_mix(x[0]);
 x[2] += k_mix(x[1]);
 x[3] += k_mix(x[2]);
 x[0] += k_mix(x[3]);
 }
}
//-----------------------------------
// Inverse mix (for testing)
//
static void unmix_128(u32 *x)
{
for(u32 i = 0; i < mix_128_rounds; i++)
 {
 x[0] -= k_mix(x[3]);
 x[3] -= k_mix(x[2]);
 x[2] -= k_mix(x[1]);
 x[1] -= k_mix(x[0]);
 }
}


void X_64(u32 *hash, u32 const* data)
{
u32 key[4]; // 128-bit key

key[0] = data[0];
key[2] = data[1];
key[1] = hash[0];
key[3] = hash[1];

mix_128(key);

XTEA_Encode(hash, key);
}

void I_64(u32 *hash, u32 const* data)
{
u32 key[4]; // 128-bit key

key[0] = data[0];
key[2] = data[1];
key[1] = hash[0];
key[3] = hash[1];

mix_128(key);

IDEA_Encode(hash, key);
}

//----------------------------------
// 128 bit hash, 64 bit data
// 
void I_128(u32 *hash, u32 const* data)
{
u32 key[4];
u32 tmp_data[2];

key[0] = data[0];
key[2] = data[1];
key[1] = hash[2];
key[3] = hash[3];

tmp_data[0] = hash[0];
tmp_data[1] = hash[1];

mix_128(key);
IDEA_Encode(tmp_data, key);

key[0] = hash[0];
key[2] = hash[1];
key[1] = data[0];
key[3] = data[1];

hash[0] ^= tmp_data[0];
hash[1] ^= tmp_data[1];

tmp_data[0] = hash[2];
tmp_data[1] = hash[3];

mix_128(key);
IDEA_Encode(tmp_data, key);

hash[2] ^= tmp_data[0];
hash[3] ^= tmp_data[1];
}

//----------------------------------
// 128 bit hash, 64 bit data
// 
void X_128(u32 *hash, u32 const* data)
{
u32 key[4];
u32 tmp_data[2];

key[0] = data[0];
key[2] = data[1];
key[1] = hash[2];
key[3] = hash[3];

tmp_data[0] = hash[0];
tmp_data[1] = hash[1];

mix_128(key);
XTEA_Encode(tmp_data, key);

key[0] = hash[0];
key[2] = hash[1];
key[1] = data[0];
key[3] = data[1];

hash[0] ^= tmp_data[0];
hash[1] ^= tmp_data[1];

tmp_data[0] = hash[2];
tmp_data[1] = hash[3];

mix_128(key);
XTEA_Encode(tmp_data, key);

hash[2] ^= tmp_data[0];
hash[3] ^= tmp_data[1];
}
//------------------------------
// 256 bit hash, 64 bit data
// This must be strong function since it is used for fingerprint calculation
//
void hash_block_256(u32 *hash, u32 const *data)
{
u32 tmp_data_x[2], tmp_data_i[2];

// Expand data

tmp_data_x[0] = tmp_data_i[0] = data[0];
tmp_data_x[1] = tmp_data_i[1] = data[1];


XTEA_Encode(tmp_data_x, hash);
IDEA_Encode(tmp_data_x, hash + 4);

IDEA_Encode(tmp_data_i, hash);
XTEA_Encode(tmp_data_i, hash + 4);

// Mix halves

X_128(hash, tmp_data_i);

hash[4] += hash[0];
hash[5] += hash[1];
hash[6] += hash[2];
hash[7] += hash[3];

I_128(hash + 4, tmp_data_x);

hash[0] += hash[4];
hash[1] += hash[5];
hash[2] += hash[6];
hash[3] += hash[7];

I_128(hash, data);

hash[4] += hash[0];
hash[5] += hash[1];
hash[6] += hash[2];
hash[7] += hash[3];

I_64(tmp_data_x, tmp_data_i);

X_128(hash + 4, tmp_data_x);
}


/* Deprecated
//----------------------------------------
// Scramble object with other object
// (Used for seed washing)
//
void scramble(void *dest, u32 dest_size_bytes, void const *src, u32 src_size_bytes)
{
u32 tmp_src[2];
u32 tmp_dest[2];


u32 max_size_bytes = (dest_size_bytes > src_size_bytes) ? dest_size_bytes : src_size_bytes;

u32 iter = (max_size_bytes+7)/8  + (dest_size_bytes+7)/8; // round up to 8 byte blocks, make two cycles through destination

u32 dest_rd_idx = 0;
u32 dest_wr_idx = 0;
u32 src_idx  = 0;

// Initialize hash

tmp_src[0] = dest_size_bytes;
tmp_src[1] = src_size_bytes;


while(iter--)
 {
 tmp_dest[0] = 0;
 tmp_dest[1] = 0;
 
 //------------    
 
 for(u32 ci = 0; ci < 8; ci++)
  {
  u32 tmp_idx = ci/4;
  u32 tmp_shift =  8*(ci & 3);
 
  // Read source, xor with previous result
 
  tmp_src[tmp_idx] ^= (u32)(((u8 const*)src)[src_idx]) << tmp_shift; // xor with previous result
  src_idx++;
  if(src_idx >= src_size_bytes) src_idx -= src_size_bytes;
  
  // Read destination

  tmp_dest[tmp_idx] |= (u32)(((u8*)dest)[dest_rd_idx]) << tmp_shift;
  dest_rd_idx++;
  if(dest_rd_idx >= dest_size_bytes) dest_rd_idx -= dest_size_bytes;
  }
 
 // Avoid degeneration

 tmp_dest[0]  ^= src_idx;
 tmp_dest[1]  ^= dest_rd_idx;

 // Hashing

 X_64(tmp_dest, tmp_src); 

 // chaining

 tmp_src[0] = tmp_dest[0];
 tmp_src[1] = tmp_dest[1]; 

 // Put result back into destination

 for(u32 ci = 0; ci < 8; ci++)
  {
  ((u8*)dest)[dest_wr_idx] = (u8)((tmp_dest[ci/4] >> (8*(ci & 3))) & 0xFF); 
  dest_wr_idx++;
  if(dest_wr_idx >= dest_size_bytes) dest_wr_idx -= dest_size_bytes;
  }

 } // iter
}
*/
//------------------------------------------------------------
// Generate hash of arbitrary data length
// Initial value of hash must be initialized 
//
static void hash_block(u32 *hash, u8 const* data, u32 data_length_bytes, void (*hash_function)(u32 *,u32 const*))
{
u32 tmp[2];

// Just for sanity

tmp[0] = 0;
tmp[1] = 0; 

for(u32 ci = 0; ci < data_length_bytes; ci++)
 {
 tmp[1] <<= 8;
 tmp[1] |= tmp[0] >> 24;
 tmp[0] <<= 8;
 tmp[0] |= data[ci];
 
 if((ci & 0x07) == 0x07) // 8 bytes loaded
  {
  hash_function(hash, tmp);
  }
 }

// Pad last block with zeroes

u32 ci = data_length_bytes & 0x07;
if(ci)
 {
 ci = 8 - ci;   
 while(ci)
  {
  ci--;
  tmp[1] <<= 8;
  tmp[1] |= tmp[0] >> 24;
  tmp[0] <<= 8;
  }
 hash_function(hash, tmp);
 }
}



//------------------------------------------------------------
// Generate 64-bit hash of arbitrary data length
// using XTEA 
// This is for MIC purpose only
// Initial value of hash must be initialized 
//
void Hash_Block_X_64(u32 *hash, u8 const* data, u32 data_length_bytes)
{
hash_block(hash, data, data_length_bytes, X_64);  
}

//------------------------------------------------------------
// Generate 256-bit hash of arbitrary data length
// Initial value of hash must be initialized 
//
void Hash_Block_256(u32 *hash, u8 const* data, u32 data_length_bytes)
{
hash_block(hash, data, data_length_bytes, hash_block_256);  
}


void Hash_Block_256(u32 *hash, u32 const* data, u32 data_length_x32)
{
u32 tmp[2] = { 0, 0 };

for(u32 ci = 0; ci < data_length_x32; ci++)
 {
 tmp[1] = tmp[0];
 tmp[0] = data[ci];
 if(ci & 1) hash_block_256(hash, tmp); // 64 bit data; 256 bit hash 
 }

if(data_length_x32 & 1)
 {
 tmp[1] = tmp[0];
 tmp[0] = 0;
 hash_block_256(hash, tmp);
 }

}


//===================== <END> =====================
