#include "types.h"
#include <string.h>
//#include <stdio.h>
#include "large_math.h"
#include "random.h"

//-------------------------------------------------------
// Z = X + Y, returns last carry (0 or 1)
//
u32 multi_add(u32 *z, u32 const * x, u32 const * y, const u32 length_x32)
{
u32 c = 0;

for(u32 ci = 0; ci < length_x32; ci++)
 {
 u64 tmp = x[ci];
 tmp += y[ci];
 tmp += c;
 z[ci] = (u32)tmp;
 c = (u32)(tmp >> 32);
 }

return c;
}

u32 multi_inc(u32 *x, const u32 length_x32)
{
u32 ret_value = 1;

for(u32 ci = 0; ci < length_x32; ci++)
 {

 x[ci]++;

 if(x[ci]) 
  {
  ret_value = 0;
  break;
  }
 }

return ret_value;
}
//-----------------
//  x = x - 1
// returns carry
//
u32 multi_dec(u32 *x, const u32 length_x32)
{
u32 ret_value = 1;

for(u32 ci = 0; ci < length_x32; ci++)
 {
 x[ci]--;
 if(x[ci] != 0xFFFFFFFF)
  {
  ret_value = 0;
  break;
  }
 }
return ret_value;
}


//-------------------------------------------------------
// Z = X - Y, returns last carry (0 or 1)
//
u32 multi_sub(u32 *z, u32 const *x, u32 const *y, const u32 length_x32)
{
u32 c = 0;

for(u32 ci = 0; ci < length_x32; ci++)
 {
 u64 tmp = x[ci];
 tmp -= y[ci];
 tmp -= c;
 z[ci] = (u32)tmp;
 c = 0x01 & (u32)(tmp >> 32);
 }

return c;
}


//------------
// Z = X * Y (unsigned)
// note Z must be twice as long as X and Y
// Note: operands should not overlap with result!

void multi_mul(u32 *z, u32 const * x, u32 const * y, const u32 length_x32)
{
zero(z, 2*length_x32); // zero the result initially

// Avoid redundant multiplications by zero

u32 y_size = length_x32;
while(y_size)
 {
 if(y[y_size-1]) break;
 y_size--;
 }

u32 x_size = length_x32;
while(x_size)
 {
 if(x[x_size-1]) break;
 x_size--;
 }



for(u32 ci = 0; ci < y_size; ci++)
 { 
 u32 m = y[ci];

 if(!m) continue; // skip zero multiplication

 u32 c = 0;

 // Partial product 

 for(u32 cj = 0; cj < x_size; cj++)
  {
  u64 p = x[cj]*(u64)m;
  p += z[ci + cj];
  p += c;
  z[ci + cj] = (u32)p;
  c = (u32)(p >> 32);
  } // cj

 // Propagate carry

 for(u32 cj = ci + x_size; cj < 2*length_x32; cj++) 
  {
  if(!c) break;   
  
  u64 tmp = z[cj];
  tmp += c;
  z[cj] = (u32)tmp;
  c = (u32)(tmp >> 32);
  } // cj

 } // ci
}
//--------------------------------------
// Faster multiplication for length = 512 bits (16 dwords) 
// (about 50% faster then general case; compiler unrolls innner loop)
// Note: result should not overlap with operands
//
void multi_mul_512(u32 *z, u32 const * x, u32 const * y)
{
const u32 length_x32 = 16;    

zero(z, 2*length_x32); // zero the result initially


for(u32 ci = 0; ci < length_x32; ci++)
 { 
 u32 m = y[ci];
 u32 c = 0;
 
 // Partial product 

 for(u32 cj = 0; cj < length_x32; cj++)
  {
  u64 p = x[cj]*(u64)m;
  p += z[ci+cj];
  p += c;
  z[ci + cj] = (u32)p;
  c = (u32)(p >> 32);
  } // cj

 // Propagate last carry 

 z[ci + length_x32] = c;
 } // ci
}


//-----------------
// returns 0 if x == y
// returns  1 if x > y
// returns -1 if x < y

s32 multi_compare(u32 const *x, u32 const * y, const u32 length_x32)
{
s32 result = 0;
u32 ci = length_x32;
    

while(ci)
 {
 ci--;

 if(x[ci] > y[ci])
  {
  result = 1;
  break;  
  }
 else if(x[ci] < y[ci])
  {
  result = -1;
  break;  
  }
 }

return result;
}

// Returns last carry

u32 multi_shift_left(u32 *x, const u32 length_x32)
{
u32 c0, c1 = 0;

for(u32 ci = 0; ci < length_x32; ci++) 
 {
 c0 = x[ci] >> 31;   
 x[ci] = (x[ci] << 1) | c1;
 c1 = c0;    
 }

return c1;
}

void multi_shift_right(u32 *x, const u32 length_x32)
{
u32 c0, c1 = 0;

u32 ci = length_x32;

while(ci)
 {
 ci--;
 c0 = x[ci] << 31;
 x[ci] = (x[ci] >> 1) | c1; 
 c1 = c0;
 }

}

//------------------------------------------------

void multi_shift_left(u32 *x, const u32 value, const u32 length_x32)
{

if(!value) return;    

u32 shft_hi = value >> 5;
u32 shft_0 = value & 0x1F;
u32 shft_1 = 32 - shft_0;

u32 ci = length_x32;

while(ci)
 {
 ci--;

 u32 tmp = 0; 

 if(ci >= shft_hi)
  {
  tmp = x[ci - shft_hi] << shft_0;

  if((ci >= (shft_hi + 1)) && (shft_1 != 32)) // damm GCC counts shift modulo 32
   {
   tmp |= x[ci - shft_hi - 1] >> shft_1;  
   }
  }

 x[ci] = tmp;
 }

}

void multi_shift_right(u32 *x, const u32 value, const u32 length_x32)
{

if(!value) return;    

u32 shft_hi = value >> 5;
u32 shft_0 = value & 0x1F;
u32 shft_1 = 32 - shft_0;

u32 ci = length_x32;


for(u32 ci = 0; ci < length_x32; ci++)
 {
 u32 tmp = 0; 
 u32 idx = ci + shft_hi;

 if(idx < length_x32)
  {
  tmp = x[idx] >> shft_0;

  idx++;

  if((idx < length_x32) && (shft_1 != 32)) // damm GCC counts shift modulo 32
   {
   tmp |= x[idx] << shft_1;  
   }
  }
 x[ci] = tmp;
 }
}



void multi_set_bit(u32 *x, const u32 bit_num)
{
x[bit_num >> 5] |= ((u32)1) << (bit_num & 0x1F);
}

void multi_clr_bit(u32 *x, const u32 bit_num)
{
x[bit_num >> 5] &= ~(((u32)1) << (bit_num & 0x1F));
}


u32 multi_get_bit(u32 const *x, const u32 bit_num)
{
return ((x[bit_num >> 5] >> (bit_num & 0x1F)) & 0x01);
}

// Exponent (number of leading zeroes)

u32 multi_exp(u32 const *x, const u32 length_x32)
{
u32 exponent = 0;    
u32 ci = length_x32;

while(ci)
 {
 ci--;
 if(x[ci]) break;
 else exponent += 32;
 }

u32 tmp = x[ci];

if(tmp)
 {

 if(!(tmp & 0xFFFF0000))
  {
  tmp <<= 16;
  exponent += 16;
  }

 while(!(tmp & 0x80000000))
  {
  tmp <<= 1;
  exponent++;  
  }
 }

return exponent;
}
//--------------------------------------------
// Extend word to double word
//
void extend_2(u32 *extended_x, u32 const *x, const u32 length_x32)
{
copy(extended_x, x, length_x32);
zero(extended_x + length_x32, length_x32);
}

void truncate_2(u32 *x, u32 const * extended_x, const u32 length_x32)
{
copy(x, extended_x, length_x32);
}

void copy(u32 *dest, u32 const * src, const u32 length_x32)
{
u32 ci = length_x32;
while(ci--) *dest++ = *src++;
}

void copy(u8 *dest, u8 const * src, const u32 length)
{
u32 ci = length;
while(ci--) *dest++ = *src++;
}

void zero(u32 *dest, const u32 length_x32)
{
u32 ci = length_x32;
while(ci--) *dest++ = 0;
}
//-------------------

u32 equal_to_zero(const u32 *x, const u32 length_x32)
{
    
for(u32 ci = 0; ci < length_x32; ci++)
 {
 if(x[ci]) return 0;
 }

return 1;
}
//------------------------

u32 equal_to_one(const u32 *x, const u32 length_x32)
{
if(x[0] != 1) return 0;    

for(u32 ci = 1; ci < length_x32; ci++)
 {
 if(x[ci]) return 0; 
 }

return 1;
}

// Returns 1 if less then two

u32 less_then_two(const u32 *x, const u32 length_x32)
{
if(x[0] > 2) return 0;

for(u32 ci = 1; ci < length_x32; ci++)
 {
 if(x[ci]) return 0;
 }

return 1;
}

//=============

MULTI_POW_MOD::MULTI_POW_MOD(u32 length_x32)
{
MULTI_POW_MOD::length = length_x32;
MULTI_POW_MOD::pwr_x  = new u32[length_x32];
MULTI_POW_MOD::tmp_prod = new u32[2*length_x32];
MULTI_POW_MOD::tmp_div_result = new u32[2*length_x32];
MULTI_POW_MOD::tmp_div_reminder = new u32[2*length_x32];
MULTI_POW_MOD::tmp_p = new u32[2*length_x32];
}

MULTI_POW_MOD::~MULTI_POW_MOD()
{
delete[] MULTI_POW_MOD::tmp_p;
delete[] MULTI_POW_MOD::tmp_div_reminder;
delete[] MULTI_POW_MOD::tmp_div_result;
delete[] MULTI_POW_MOD::tmp_prod;
delete[] MULTI_POW_MOD::pwr_x;
}

//----------------------------------------------
// result = (X ^ Y) mod P
//  
void MULTI_POW_MOD::multi_pow_mod(u32 *result, const u32 *x, const u32 *y, const u32 *p)
{
// Initialize pwr_x to X

copy(pwr_x, x, length);  

// Initialize result to 1

zero(result, length);
result[0] = 1;


for(u32 n = 0; n < (length * 32); n++)
 { 
 if(multi_get_bit(y,n))
  {
  multi_mul(tmp_prod, result, pwr_x, length);
  extend_2(tmp_p, p, length);
  multi_div(tmp_div_result, tmp_div_reminder, tmp_prod, tmp_p, 2*length);
  truncate_2(result, tmp_div_reminder, length);  
  }

 // Calculate power(x) modulo p    

 multi_mul(tmp_prod, pwr_x, pwr_x, length);
 extend_2(tmp_p, p, length);
 multi_div(tmp_div_result, tmp_div_reminder, tmp_prod, tmp_p, 2*length);
 truncate_2(pwr_x, tmp_div_reminder, length);
 }
}

//----------------------------
// X/Y
// division by zero is not checked
// x and y are destroyed
//
void multi_div(u32 *result,u32 *remainder, u32 *x, u32 *y, u32 length_x32)
{
zero(result,  length_x32);
zero(remainder, length_x32);
    

if(multi_compare(x,y,length_x32) < 0)
 {
 copy(remainder, x, length_x32);
 return;
 }

// Normalize divisor

u32 e = multi_exp(y, length_x32) - multi_exp(x, length_x32);

multi_shift_left(y, e, length_x32);


// Division loop 

do {
   u32 c = multi_sub(remainder, x, y, length_x32);
   if(!c) 
    {
    multi_set_bit(result, e);
    copy(x, remainder, length_x32);
    }
   multi_shift_right(y, length_x32);
   } while(e--);

copy(remainder, x, length_x32);
}

MULTI_DIV::MULTI_DIV(u32 length_x32)
{
length = length_x32;    
MULTI_DIV::x = new u32[length];
MULTI_DIV::y = new u32[length];
}

MULTI_DIV::~MULTI_DIV()
{
MULTI_DIV::reset();    
delete[] MULTI_DIV::y;
delete[] MULTI_DIV::x;
}

void MULTI_DIV::reset()
{
cheap_random_number(MULTI_DIV::x, length);
cheap_random_number(MULTI_DIV::y, length);
}

void MULTI_DIV::multi_div(u32 *quotient, u32 *remainder, u32 const * xx, u32 const * yy)
{
copy(x, xx, length);
copy(y, yy, length);
::multi_div(quotient, remainder, x, y, length);
}

SIGNED_MULTI_DIV::SIGNED_MULTI_DIV(u32 length_x32)
{
length = length_x32;    
SIGNED_MULTI_DIV::x = new u32[length_x32];
SIGNED_MULTI_DIV::y = new u32[length_x32];
}

SIGNED_MULTI_DIV::~SIGNED_MULTI_DIV()
{
SIGNED_MULTI_DIV::reset();    
delete[] SIGNED_MULTI_DIV::y;
delete[] SIGNED_MULTI_DIV::x;
}

void SIGNED_MULTI_DIV::reset()
{
cheap_random_number(x, length);
cheap_random_number(y, length);
}


//-------------------------------
// Remainder is signed 
//
void SIGNED_MULTI_DIV::signed_multi_div(u32 *quotient, u32 *remainder, u32 const * xx, u32 const * yy)
{
copy(x, xx, length);
copy(y, yy, length);

int sx = 0;
int sy = 0;

if(x[length-1] & 0x80000000)
 {
 multi_negate(x, length); 
 sx = 1;
 }

if(y[length-1] & 0x80000000)
 {
 multi_negate(y, length);
 sy = 1;
 }

::multi_div(quotient, remainder, x, y, length);

if(sx ^ sy) multi_negate(quotient, length);

if(sx) multi_negate(remainder, length);
}

SIGNED_MULTI_MUL::SIGNED_MULTI_MUL(u32 length_x32)
{
length = length_x32;    
SIGNED_MULTI_MUL::x = new u32[length_x32];
SIGNED_MULTI_MUL::y = new u32[length_x32];
}

SIGNED_MULTI_MUL::~SIGNED_MULTI_MUL()
{
SIGNED_MULTI_MUL::reset();    
delete[] SIGNED_MULTI_MUL::y;
delete[] SIGNED_MULTI_MUL::x;
}

void SIGNED_MULTI_MUL::reset()
{
cheap_random_number(x, length);
cheap_random_number(y, length);
}

//-------------------------------
// Remainder is signed 
//
void SIGNED_MULTI_MUL::signed_multi_mul(u32 *product, u32 const * xx, u32 const * yy)
{
copy(x, xx, length);
copy(y, yy, length);

int s = 0;

if(x[length-1] & 0x80000000)
 {
 multi_negate(x, length); 
 s = 1;
 }

if(y[length-1] & 0x80000000)
 {
 multi_negate(y, length);
 s ^= 1;
 }

::multi_mul(product, x, y, length);

if(s) multi_negate(product, 2*length);
}

//---------------------
// x = -x
//
void multi_negate(u32 *x, u32 length_x32)
{
for(u32 ci = 0; ci < length_x32; ci++) x[ci] ^= 0xFFFFFFFF;

multi_inc(x, length_x32);
}

//----------------------

MODULO_MUL::MODULO_MUL(u32 length_x32)
{
len = length_x32;
r = new u32[2*len];
q = new u32[2*len];
n = new u32[2*len];
acc = new u32[2*len];
}

MODULO_MUL::~MODULO_MUL()
{
delete[] acc;    
delete[] n;    
delete[] q;
delete[] r;
}

void MODULO_MUL::modulo_mul(u32 *result, u32 const *x, u32 const *y, u32 const* p)
{
multi_mul(acc, x, y, len);
copy(n,p,len);
zero(n+len, len);
multi_div(q, r, acc, n, 2*len); 
copy(result, r, len);
}



//=============== <END> =================
