#include "pkf.h"
#include <string.h>


static const s64 pkf_p = 0x7e012aab8b1f14f3; // Random 63-bit safe prime 
static const u32 pkf_mem_size_dwords = 262144; // 1 MB; must be power of 2
static const u32 shuffle_iter = 3*pkf_mem_size_dwords; // ~ 20 iter/s

//----
static const u32 pkf_idx_mask = pkf_mem_size_dwords - 1;


u32 PKF::mix(u32 *crc, const u32 x)
{
crc[0] += x;

for(u32 i = 0; i < 7; i++)
 {   
 u32 t = crc[0];
 t ^= t << 5;
 crc[0] = crc[1];
 crc[1] = crc[2];
 u32 w = crc[2] = crc[3];
 crc[3] = (w ^ (w >> 29)) ^ (t ^ (t >> 12));
 }

return crc[3];
}


//----------------------
// Find x^-1 mod y
// i.e. (x x^-1) mod y = 1
// Intentionally slow 
//
s64 PKF::inverse(const s64 modulo, s64 value)
{
// 0 -> modulo
// 1 -> 1
// ... -> ...
// modulo -> 0

if(!value) return modulo; 

value = value % modulo;    

if(value <= 1) return value; 

s64 x[3],y[3],r[3];

x[0] = 1; x[1] = 0;
y[0] = 0; y[1] = 1;
r[0] = modulo; r[1] = value;

for(;;)
 {
 s64 q = r[0] / r[1];
 r[2] =  r[0] % r[1];

 if(!r[2]) break;
 
 x[2] = x[0] - q*x[1];
 y[2] = y[0] - q*y[1];

 x[0] = x[1]; x[1] = x[2];
 y[0] = y[1]; y[1] = y[2];
 r[0] = r[1]; r[1] = r[2];
 }

if(y[1] < 0) y[1] += modulo;

return y[1];
}


void PKF::init()
{
for(u32 i = 0; i < pkf_mem_size_dwords; i++) pkf_mem[i] = i;
}

PKF::PKF()
{
pkf_mem = new u32[pkf_mem_size_dwords];
PKF::init();
}

PKF::~PKF()
{
PKF::init();
delete[] pkf_mem;
}




void PKF::get_key(void *key, const u32 key_size_bytes,
                  void const *salt, const u32 salt_size_bytes, 
                  void const *passwd, const u32 passwd_size_bytes)
{
u32 crc[4];
memset(crc, 0xFF, sizeof(crc));
memset(key, 0, key_size_bytes);

 
// Shuffle

u32 xp = 0, xs = 0, k0 = (u32)pkf_p, k1 = (u32)(pkf_p >> 32);
u32 salt_idx = 0, passwd_idx = 0;
u32 j = 0;


for(u32 n = 0; n < shuffle_iter; n++)
 {

 // Read 4 bytes of salt and password
  
 for(u32 i = 0; i < sizeof(u32); i++)
  {
  xs = (xs << 8) |  ((u8*)salt)[salt_idx];
  salt_idx++;
  if(salt_idx >= salt_size_bytes) salt_idx = 0;

  xp = (xp << 8) |  ((u8*)passwd)[passwd_idx];
  passwd_idx++;
  if(passwd_idx >= passwd_idx) passwd_idx = 0;
  }

 // Hash into state

 u32 x = PKF::mix(crc, xs ^ k0);
 u32 y = PKF::mix(crc, xp ^ k1);

 // Shuffle

 u32 i = n & pkf_idx_mask;
 u32 Si = pkf_mem[i];   
 j = (j + Si + x) & pkf_idx_mask;
 u32 Sj = pkf_mem[j];
 pkf_mem[i] = Sj;
 
 if(x > 0xF0000000) 
  {
  u64 z = (((u64)Si) << 32) + Sj;
  if(z > (u64)pkf_p) z = z % (u64)pkf_p;
 
  z = (u64)PKF::inverse(pkf_p, (s64)z);
  k0 = (u32)z;
  k1 = (u32)(z >> 32);
  }

 Si = Sj ^ (Si + y);
 Si = (Si << 11) | (Si >> 21);
   
 pkf_mem[j] = Si;

 y = PKF::mix(crc, Si ^ n);
 u32 key_idx = y % key_size_bytes; 
 y += y >> 16;
 y ^= y >> 8;
 ((u8*)key)[key_idx] += y;
 }

// Cleanup

memset(crc, 0xFF, sizeof(crc));
PKF::init();
}

