#include "kademlia.h"
#include "network.h"
#include "resource.h"
#include "users.h"
#include "main.h"
#include <string.h>
#include <time.h>
//---------------------------------------------------
static const u32 PROT_PING = 0x01;
static const u32 PROT_FIND_NODE = 0x02;
static const u32 PROT_REPLY = 0x80000000;
//---------------------------------------------------
KADEMLIA::KADEMLIA(const NODE *my_node, HWND mWnd)
{
	mainWnd = mWnd;
	memcpy(&me,my_node,sizeof(NODE));               // copy our node information
	memset(grp,0,sizeof(grp));                      // clear table
	InitializeCriticalSection(&cs);
	eStopThreads = CreateEvent(NULL,TRUE,FALSE,NULL);
	sysThread = NULL;
	inFinding = false;
	memset(findingDHT,0,sizeof(findingDHT));              // clear table
	eFoundNode = CreateEvent(NULL,FALSE,FALSE,NULL);
	inFinding = false;
	inBootstrapping = false;
	sysThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)systemProc,this,0,NULL);
}
//---------------------------------------------------
KADEMLIA::~KADEMLIA()
{
	if (sysThread) stop(); // stop thread if running
	DeleteCriticalSection(&cs);
	CloseHandle(eStopThreads);
	for (u32 i = 0; i < NETWORK_ID_LENGTH*32; i++) // clean all DHT chains in each bucket
	{
		DHT *d1, *d2;
		for (d1 = d2 = grp[i].dht; d1; d1 = d2)
		{
			d2 = d1->next;
			delete d1;
		}
	}
}
//---------------------------------------------------
bool isIdIdentical(const u32* id1, const u32* id2)
{
	if (!id1 || !id2) return false;
    for (int i = 0; i < NETWORK_ID_LENGTH; i++)
    {
        if (id1[i] != id2[i]) return false;
    }
    return true;
}
//---------------------------------------------------
static bool distanceCloser(const u32* ref_id, const u32* new_id, const u32* old_id)
{
    for (int i = 0; i < NETWORK_ID_LENGTH; i++)
    {
        u32 new_d = ref_id[i] ^ new_id[i];
        u32 old_d = ref_id[i] ^ old_id[i];
        if (new_d < old_d) return true;
        if (new_d > old_d) return false;
    }
    return false;
}
//---------------------------------------------------
static int calcBucketNum(const u32* id1, const u32* id2)
{
    u32 d[NETWORK_ID_LENGTH];
    int i;
    for (i = 0; i < NETWORK_ID_LENGTH; i++) d[i] = id1[i] ^ id2[i]; // distance calculation
    for (i = NETWORK_ID_LENGTH*32-1; i >= 0; i--)
    {
        if (d[NETWORK_ID_LENGTH-1-(i>>5)] & (1<<(i&0x1F))) break; // found bucket number
    }
    return i;
}
//---------------------------------------------------
static void moveToHead(BUCKET* g, DHT* cur_dht, DHT* prev_dht)
{
    if (cur_dht != g->dht) // if it's not already the first item
    {
        prev_dht->next = cur_dht->next; // remove current DHT from the chain
        prev_dht = g->dht;				// remember first DHT
        g->dht = cur_dht;				// now current DHT is the first one (head)
        g->dht->next = prev_dht;		// point to previous first DHT
    }
}
//---------------------------------------------------
static void moveFromWaitingList(BUCKET* g, DHT* cur_dht, DHT* prev_dht)
{
    memcpy(&cur_dht->node,&g->waiting_list[0],sizeof(NODE));
    cur_dht->firstPingTime = cur_dht->lastPingTime = clock();
    cur_dht->flags = 0;
    cur_dht->ping_count = 0;
    moveToHead(g, cur_dht, prev_dht);
    // shift waiting list
    memmove(&g->waiting_list[0],&g->waiting_list[1],(KADEMLIA_WAITING_MAX-1)*sizeof(NODE));
    memset(&g->waiting_list[KADEMLIA_WAITING_MAX-1],0,sizeof(NODE));
}
//---------------------------------------------------
DWORD WINAPI systemProc(KADEMLIA* kad)
{
    DHT *cur_dht, *prev_dht;
    while (WAIT_TIMEOUT == WaitForSingleObject(kad->eStopThreads,1000)) // processing each second
    {
        for (int i = 0; i < NETWORK_ID_LENGTH*32; i++)
        {
            EnterCriticalSection(&kad->cs);
            while (kad->grp[i].waiting_list[0].ip) // if someone waits adding
            {
                // try to find someone lost
                prev_dht = NULL;
                for (cur_dht = kad->grp[i].dht; cur_dht; cur_dht = cur_dht->next)
                {
                    if (cur_dht->flags & KAD_DHT_FLAG_LOST)
                    {
                        moveFromWaitingList(&kad->grp[i],cur_dht,prev_dht);
                        break;
                    }
                    prev_dht = cur_dht;
                }
                if (!cur_dht) // no one was lost, ping sequentially
                {
                    prev_dht = NULL;
                    for (cur_dht = kad->grp[i].dht; cur_dht; cur_dht = cur_dht->next)
                    {
                        if (!(cur_dht->flags & KAD_DHT_FLAG_PINGED)) // no pinged item, ping it
                        {
                            cur_dht->flags |=  KAD_DHT_FLAG_PINGED;
                            cur_dht->flags &= ~KAD_DHT_FLAG_REPLIED;
                            cur_dht->ping_count++;
                            kad->sendPing(&cur_dht->node,false);
                            break;
                        }
                        else if (!(cur_dht->flags & KAD_DHT_FLAG_REPLIED)) // if pinged but not replied
                        {
                            if (cur_dht->ping_count > KADEMLIA_PING_MAX) // lost item
                            {
                                moveFromWaitingList(&kad->grp[i],cur_dht,prev_dht);
                            }
                            else
                            {
                                cur_dht->ping_count++;
                                kad->sendPing(&cur_dht->node,false);
                            }
                            break;
                        }
                        prev_dht = cur_dht;
                    }
                    if (!cur_dht) // all table checked, everyone alive, clear all flags and waiting list
                    {
                        for (cur_dht = kad->grp[i].dht; cur_dht; cur_dht = cur_dht->next)
                        {
                            cur_dht->flags = 0;
                            cur_dht->ping_count = 0;
                        }
                        memset(&kad->grp[i].waiting_list[0],0,KADEMLIA_WAITING_MAX*sizeof(NODE));
                        break;
                    }
                }
            }
            LeaveCriticalSection(&kad->cs);
        }
        if (kad->inBootstrapping)
        {
            for (int i = 0; i < NETWORK_ID_LENGTH*32; i++)
            {
                DHT* d;
                EnterCriticalSection(&kad->cs); // deadlock preventing
                for (d=kad->grp[i].dht;d;d=d->next)
                {
                    kad->sendRequestID(&d->node,kad->me.id);
                }
                LeaveCriticalSection(&kad->cs);
            }
        }
    }
    return 0;
}
//---------------------------------------------------
void KADEMLIA::start()
{
	// bootstraping
	inFinding = true;
	inBootstrapping = true;
}
//---------------------------------------------------
void KADEMLIA::stop()
{
	if (!sysThread) return; // if already stopped
	SetEvent(eStopThreads);
	WaitForSingleObject(sysThread,INFINITE);
	CloseHandle(sysThread);
	sysThread = NULL;
}
//---------------------------------------------------
void KADEMLIA::sendRequestID(const NODE* nd, const u32 *id)
{
	u32 data[NETWORK_ID_LENGTH+1];
	data[0] = PROT_FIND_NODE;
	memcpy(&data[1],id,NETWORK_ID_LENGTH*sizeof(u32));
 
	PACKET_FOR_SEND pack;
	pack.buffer = data;
	pack.buf_len = (NETWORK_ID_LENGTH+1)*sizeof(u32);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_KADEMLIA;
	getNet()->send_packet(&pack);
}
//---------------------------------------------------
void KADEMLIA::sendReplyID(const NODE* nd, const u32 *id)
{
	u32 data[2+(NETWORK_ID_LENGTH+2)*KADEMLIA_K]; // maximum packet size
	data[0] = PROT_FIND_NODE|PROT_REPLY;
	u32 k = fillBucket(id, &data[2]);
	data[1] = k;
 
	PACKET_FOR_SEND pack;
	pack.buffer = data;
	pack.buf_len = (2+k*(NETWORK_ID_LENGTH+2))*sizeof(u32);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_KADEMLIA;
	getNet()->send_packet(&pack);
}
//---------------------------------------------------
void KADEMLIA::sendPing(const NODE* nd, bool reply)
{
	u32 data[1];
	data[0] = PROT_PING;
	if (reply) data[0] |= PROT_REPLY;
 
	PACKET_FOR_SEND pack;
	pack.buffer = data;
	pack.buf_len = sizeof(u32);
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_KADEMLIA;
	getNet()->send_packet(&pack);
}
//---------------------------------------------------
void KADEMLIA::updateNode(const NODE* nd, bool trueReply, USERS* usr)
{
	u32 i;
	int bucket = calcBucketNum(nd->id,me.id);
	if ((bucket < 0) || (bucket >= NETWORK_ID_LENGTH*32)) return; // myself ID or out of range, it's very strange!!!
	BUCKET* g = &grp[bucket];
	DHT *cur_dht, *prev_dht = NULL;
	EnterCriticalSection(&cs);
	// try to find such node in table
	i = 0;
	for (cur_dht = g->dht; cur_dht; cur_dht = cur_dht->next)
	{
		if (isIdIdentical(nd->id,cur_dht->node.id)) // this node already exists in DHT
		{
			if (trueReply)
			{
				cur_dht->node.ip = nd->ip;   // port and IP address might change
				cur_dht->node.port = nd->port;
                cur_dht->lastPingTime = clock();
				cur_dht->flags |= KAD_DHT_FLAG_REPLIED;
				cur_dht->flags &= ~KAD_DHT_FLAG_LOST;
				cur_dht->ping_count = 0;
				moveToHead(g, cur_dht, prev_dht);
			}
			break;
		}
		prev_dht = cur_dht; // for normal chain processing
		i++;
	}
	if (!cur_dht) // node wasn't found (unknown node)
	{
		if (i < KADEMLIA_K) // empty space exists
		{
			if (prev_dht) // if any item exists
			{
				prev_dht->next = new DHT;
				cur_dht = prev_dht->next;
			}
			else // first element in table
			{
				g->dht = new DHT;
				cur_dht = g->dht;
			}
			if (cur_dht) // if memory allocated
			{
				memcpy(&cur_dht->node,nd,sizeof(NODE));
                cur_dht->firstPingTime = cur_dht->lastPingTime = clock();
				cur_dht->flags = 0;
				cur_dht->ping_count = 0;
				cur_dht->next = NULL;
			}
		}
		else if (trueReply) // there is no free space, put to waiting list
		{
			for (i = 0; i < KADEMLIA_WAITING_MAX; i++)
			{
				if (!g->waiting_list[i].ip) // found empty space in waiting list
				{
					memcpy(&g->waiting_list[i],nd,sizeof(NODE));
					break;
				}
			}
		}
	}
	LeaveCriticalSection(&cs);
	if (trueReply && usr) usr->update_node(nd);
}
//---------------------------------------------------
void KADEMLIA::packetReceived(u32* pack, u32 pack_len, const NODE* nd, USERS* usr)
{
	if (pack_len >= sizeof(u32)) // contains actual data
	{
		switch (pack[0] & ~PROT_REPLY) // message code
		{
		case PROT_PING:
			if (!(pack[0] & PROT_REPLY))
			{
				sendPing(nd,true);
			}
			updateNode(nd,true,usr);
            //inBootstrapping = false;
			break;
		case PROT_FIND_NODE:
			if ((pack_len >= (NETWORK_ID_LENGTH+1)*sizeof(u32)) && !(pack[0] & PROT_REPLY)) // if correct request
			{
				sendReplyID(nd,&pack[1]);
			}
			else if ((pack[0] & PROT_REPLY) && ((u32)pack_len >= (2)*sizeof(u32)) && 
				((u32)pack_len >= (2+pack[NETWORK_ID_LENGTH+1]*(NETWORK_ID_LENGTH+2))*sizeof(u32))) // if correct reply
			{
				findindReply(&pack[1], pack_len-(1)*sizeof(u32), nd);
				// update nodes from the reply
				for (u32 i = 0; i < pack[1]; i++)
				{
					NODE rep_nd;
					memcpy(rep_nd.id,&pack[2+(NETWORK_ID_LENGTH+2)*i],NETWORK_ID_LENGTH*sizeof(u32));
					rep_nd.ip = pack[2+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH];
					rep_nd.port = (u16)pack[2+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH+1];
					updateNode(&rep_nd,false,NULL);
				}
			}
			updateNode(nd,true,usr);
            if (inBootstrapping) // delete zero DHT from table
            {
                u32 zero[NETWORK_ID_LENGTH] = {0};
                int bucket = calcBucketNum(zero,me.id);
                if ((bucket >= 0) && (bucket < NETWORK_ID_LENGTH*32))
                {
                    DHT *cur_dht, *prev_dht = NULL;
                    EnterCriticalSection(&cs);
                    for (prev_dht = cur_dht = grp[bucket].dht; cur_dht; cur_dht = cur_dht->next)
                    {
                        if (isIdIdentical(zero,cur_dht->node.id)) // this node already exists in DHT
                        {
                            if (cur_dht == prev_dht)
                            {
                                if (cur_dht->next) grp[bucket].dht = cur_dht->next;
                                else grp[bucket].dht = NULL;
                            }
                            else
                                prev_dht->next = cur_dht->next;
                            delete cur_dht;
                            break;
                        }
                        prev_dht = cur_dht;
                    }
                    LeaveCriticalSection(&cs);
                }
                inBootstrapping = false;
            }
			break;
		default: // something wrong!!!
			break;
		}
	}
}
//---------------------------------------------------
bool KADEMLIA::findNode(const u32 *id, NODE *nd)
{
	int bucket = calcBucketNum(id,me.id);
	int k,a,buck;
	DHT *dht;
	if (bucket < 0) // is't me!!!
	{
		memcpy(nd,&me,sizeof(NODE));
		return true;
	}
	EnterCriticalSection(&cs);
	for (dht = grp[bucket].dht; dht; dht = dht->next)
	{
		if (isIdIdentical(dht->node.id,id))
		{
			memcpy(nd,&dht->node,sizeof(NODE));
			LeaveCriticalSection(&cs);
			return true;
		}
	}
	if (inFinding) // another finding process is in running
	{
		LeaveCriticalSection(&cs);
		return false;
	}
	// start to request for such nodec
	k = 0;
	for (dht = grp[bucket].dht; dht; dht = dht->next)
	{
		memcpy(&findingDHT[k],dht,sizeof(DHT));
		findingDHT[k].flags = 0;
		k++;
	}
	for (buck = bucket-1; (k < KADEMLIA_K) && (buck >= 0); buck--)
	{
		for (dht = grp[buck].dht; dht && (k < KADEMLIA_K); dht = dht->next)
		{
			memcpy(&findingDHT[k],dht,sizeof(DHT));
			findingDHT[k].flags = 0;
			k++;
		}
	}
	for (buck = bucket+1; (k < KADEMLIA_K) && (buck < 32*NETWORK_ID_LENGTH); buck++)
	{
		for (dht = grp[buck].dht; dht && (k < KADEMLIA_K); dht = dht->next)
		{
			memcpy(&findingDHT[k],dht,sizeof(DHT));
			findingDHT[k].flags = 0;
			k++;
		}
	}
	memset(&findingDHT[k],0,(KADEMLIA_K-k)*sizeof(DHT));
	inFinding = true;
	memcpy(findigNode.id,nd->id,sizeof(u32)*NETWORK_ID_LENGTH);
	ResetEvent(eFoundNode);
	LeaveCriticalSection(&cs);
	for (a = 0; a < KADEMLIA_A; a++)
	{
		requestID();
	}
	if (WAIT_OBJECT_0 == WaitForSingleObject(eFoundNode,5000))
	{
		EnterCriticalSection(&cs);
		memcpy(nd,&findigNode,sizeof(NODE));
		inFinding = false;
		LeaveCriticalSection(&cs);
		return true;
	}
	EnterCriticalSection(&cs);
	inFinding = false;
	LeaveCriticalSection(&cs);
	return false;
}
//---------------------------------------------------
void KADEMLIA::requestID(void)
{
	EnterCriticalSection(&cs);
	if (!inFinding)
	{
		LeaveCriticalSection(&cs);
		return;
	}
	// find closest node
	int i, inx = -1;
	for (i = 0; i < KADEMLIA_K; i++)
	{
		if (findingDHT[i].node.ip && !(findingDHT[i].flags & KAD_DHT_FLAG_PINGED))
		{
			if ((inx < 0) || distanceCloser(findigNode.id,findingDHT[i].node.id,findingDHT[inx].node.id))
			{
				inx = i;
			}
		}
	}
	if (inx >= 0)
	{
		sendRequestID(&findingDHT[inx].node,findingDHT[inx].node.id);
		findingDHT[inx].flags = KAD_DHT_FLAG_PINGED;
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void KADEMLIA::findindReply(const u32* pack, u32 pack_len, const NODE* nd)
{
	u32 k = pack[0];
	u32 i,j;
	EnterCriticalSection(&cs);
	if (!inFinding)
	{
		LeaveCriticalSection(&cs);
		return;
	}
	// is the sender what we are looking for?
	if (isIdIdentical(nd->id,findigNode.id))
	{
		memcpy(&findigNode,nd,sizeof(NODE));
		SetEvent(eFoundNode);
		LeaveCriticalSection(&cs);
		return;
	}
	// finding through reply
	for (i = 0; i < k; i++)
	{
		if (isIdIdentical(&pack[1+(NETWORK_ID_LENGTH+2)*i],findigNode.id))
		{
			findigNode.ip = pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH];
			findigNode.port = (u16)pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH+1];
			SetEvent(eFoundNode);
			LeaveCriticalSection(&cs);
			return;
		}
	}
	// updating closest DHT
	for (i = 0; i < k; i++)
	{
		for (j = 0; j < KADEMLIA_K; j++)
		{
			if (!findingDHT[j].node.ip) // add new item into search list
			{
				memcpy(findingDHT[j].node.id,&pack[1+(NETWORK_ID_LENGTH+2)*i],NETWORK_ID_LENGTH*sizeof(u32));
				findingDHT[j].node.ip = pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH];
				findingDHT[j].node.port = (u16)pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH+1];
				findingDHT[j].flags = 0;
				break;
			}
			else if (isIdIdentical(&pack[1+(NETWORK_ID_LENGTH+2)*i],findingDHT[j].node.id)) // already exists is list
			{
				break;
			}
		}
		if (j == KADEMLIA_K) // unknown item
		{
			// try to find furthest node in list
			int farthest = -1;
			for (j = 0; j < KADEMLIA_K; j++)
			{
				if (!findingDHT[j].node.ip)
				{
					break;
				}
				else if ((farthest < 0) || !distanceCloser(findigNode.id,findingDHT[j].node.id,findingDHT[farthest].node.id))
				{
					farthest = j;
				}
			}
			if ((farthest >= 0) && distanceCloser(findigNode.id,&pack[1+(NETWORK_ID_LENGTH+2)*i],findingDHT[farthest].node.id)) // furthest found and it's farther than new
			{
				memcpy(findingDHT[farthest].node.id,&pack[1+(NETWORK_ID_LENGTH+2)*i],NETWORK_ID_LENGTH*sizeof(u32));
				findingDHT[farthest].node.ip = pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH];
				findingDHT[farthest].node.port = (u16)pack[1+(NETWORK_ID_LENGTH+2)*i+NETWORK_ID_LENGTH+1];
				findingDHT[farthest].flags = 0;
			}
		}
	}
	LeaveCriticalSection(&cs);
	for (int a = 0; a < KADEMLIA_A; a++)
	{
		requestID();
	}
}
//---------------------------------------------------
u32 KADEMLIA::fillBucket(const u32* id, u32* buf)
{
	int bucket = calcBucketNum(id,me.id);
	u32 k;
	int buck;
	DHT* dht;
	if (bucket < 0) // is't me, nothing to send
	{
		return 0;
	}
	// start to request for such nodes
	k = 0;
	for (dht = grp[bucket].dht; dht; dht = dht->next)
	{
		memcpy(&buf[k*(NETWORK_ID_LENGTH+2)],&dht->node.id,NETWORK_ID_LENGTH*sizeof(u32));
		buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH] = dht->node.ip;
		buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH+1] = dht->node.port;
		k++;
	}
	// closer distance
	for (buck = bucket-1; (k < KADEMLIA_K) && (buck >= 0); buck--)
	{
		for (dht = grp[buck].dht; dht && (k < KADEMLIA_K); dht = dht->next)
		{
			memcpy(&buf[k*(NETWORK_ID_LENGTH+2)],&dht->node.id,NETWORK_ID_LENGTH*sizeof(u32));
			buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH] = dht->node.ip;
			buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH+1] = dht->node.port;
			k++;
		}
	}
	// longer distance
	for (buck = bucket+1; (k < KADEMLIA_K) && (buck < 32*NETWORK_ID_LENGTH); buck++)
	{
		for (dht = grp[buck].dht; dht && (k < KADEMLIA_K); dht = dht->next)
		{
			memcpy(&buf[k*(NETWORK_ID_LENGTH+2)],&dht->node.id,NETWORK_ID_LENGTH*sizeof(u32));
			buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH] = dht->node.ip;
			buf[k*(NETWORK_ID_LENGTH+2)+NETWORK_ID_LENGTH+1] = dht->node.port;
			k++;
		}
	}
	return k;
}
//---------------------------------------------------
bool isIdNonZero(const u32* id)
{
    u32 cnt;
    for (cnt = 0; cnt < NETWORK_ID_LENGTH; cnt++)
    {
        if (id[cnt]) return true;
    }
    return false;
}
//---------------------------------------------------
